/* Copyright (c) 2001 - 2010, Markus Strehle, SAP AG
 *
 * MIT License, http://www.opensource.org/licenses/mit-license.php
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 */

/*source--------------------------------------------------------------*/
/*                                                                    */
/* Description:                                                       */
/* ============                                                       */
/*                                                                    */
/*                                                                    */
/*                                                                    */
/*  Author:    Markus Strehle, SAP AG (mailto:markus.strehle@sap.com) */
/*  Reviewer:                                                         */
/*  Version:       0.96                                               */
/*                                                                    */
/*  Created:                                                          */
/*    20 June 2005  Markus Strehle                                    */
/*                                                                    */
/*  Modified:                                                         */
/*    25 Mar  2011  Markus Strehle                                    */
/*                  Adaption for ClamAV 0.97                          */
/*    01 Nov  2010  Markus Strehle                                    */
/*                  Adaption for ClamAV 0.96                          */
/*    05 Aug  2009  Markus Strehle                                    */
/*                  Adaption for ClamAV 0.95                          */
/*    15 Aug  2008  Markus Strehle                                    */
/*                  (many changes)                                    */
/*                                                                    */
/*  Remarks:                                                          */
/*      Modifications are done for NW-VSI Version 1.00                */
/*                                                                    */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/*                                                                    */
/* --- public functions ----------------------------------------------*/
/*                                                                    */
/*      VsaStartup                                                    */
/*      VsaInit                                                       */
/*      VsaGetConfig                                                  */
/*      VsaScan                                                       */
/*      VsaReleaseScan                                                */
/*      VsaEnd                                                        */
/*      VsaCleanup                                                    */
/*                                                                    */
/* --- private functions ---------------------------------------------*/
/*      setScanError                                                  */
/*      vsaSetScanConfig                                              */
/*      registerCallback                                              */
/*      freevirusinfo                                                 */
/*      freevirusinfo2                                                */
/*      freescanerror                                                 */
/*      freescanerror2                                                */
/*      freeVSA_INIT                                                  */
/*      freeVSA_CONFIG                                                */
/*      getFileSize                                                   */
/*                                                                    */
/**********************************************************************/
/*--------------------------------------------------------------------*/
/* Adapter defines                                                    */
/*--------------------------------------------------------------------*/
#ifdef SAPwithUNICODE
#undef SAPwithUNICODE
#undef UNICODE 
#undef _UNICODE
#endif

/*--------------------------------------------------------------------*/
/* Own includes                                                       */
/*--------------------------------------------------------------------*/
#include "vsaxxtyp.h"
#include "vsclam.h"
/*--------------------------------------------------------------------*/
/* ClamAV includes                                                    */
#include "clamav.h"

/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/
/* system includes (OS-dependent)                                     */
/*--------------------------------------------------------------------*/
#ifdef _WIN32
#ifndef WIN32_MEAN_AND_LEAN
#define WIN32_MEAN_AND_LEAN
#include <windows.h>
#endif
#endif

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h> 

/*--------------------------------------------------------------------*/
/* static globals                                                     */
/*--------------------------------------------------------------------*/
static Bool           bgInit                =   FALSE;
static size_t         lgRefCounter          =   0;
static const char        builddate[]        =   "[DATE]CLAMSAP: " __DATE__ ", " __TIME__ ;
#ifdef __cplusplus
static const char        version[]          =   "@[CPP]CLAMSAP: " VSA_ADAPTER_VERSION;
#else
static const char        version[]          =   "@[ C ]CLAMSAP: " VSA_ADAPTER_VERSION;
#endif

/* all known messages, provided also in VSA_CONFIG */
static const UInt     uigVS_SAP_ALL = VS_M_ERROR              |
                                      VS_M_ABORTSCAN          |
                                      VS_M_VIRUS              |
                                      VS_M_CLEAN              |
                                      VS_M_NOTSCANNED         |
                                      VS_M_OBJECTFOUND;

/*--------------------------------------------------------------------*/
/* helper functions                                                   */
/*--------------------------------------------------------------------*/
static VSA_RC setScanError(UInt            uiJobID,
                           PChar           pszObjectName,
                           size_t          lObjectSize,
                           Int             iErrorRC,
                           PChar           pszErrorText,
                           PPVSA_SCANERROR pp_scanerror);
/*
 *  Assign the VSA_OPTPARAM values to
 *  internal configuration.
 */
static VSA_RC vsaSetScanConfig(VSA_SCANPARAM *,
                               VSA_OPTPARAMS *, 
                               USRDATA *,
                               struct cl_engine *engine
                               );
static VSA_RC vsaSetInitConfig(VSA_INITPARAMS *,
                               INITDATA *
                               );

static VSA_RC registerCallback( VSA_CALLBACK *, 
                                USRDATA *
                              );

static void freevirusinfo(VSA_VIRUSINFO **);
static void freevirusinfo2(VSA_VIRUSINFO *);
static void freescanerror(VSA_SCANERROR **);
static void freescanerror2(VSA_SCANERROR *);

static void freeVSA_INIT(  VSA_INIT **);
static void freeVSA_CONFIG(VSA_CONFIG **);

static VSA_RC getFileSize(Char *, size_t *);


/*--------------------------------------------------------------------*/
/* VSA public functions                                               */
/*--------------------------------------------------------------------*/
/**********************************************************************
 *  VsaStartup()
 *
 *  Description:
 *     Global initialization of the adapter. This function will be called
 *     once after loading the VSA.
 *
 *  Returncodes: 
 *  VSA_OK                   |      Success
 *  VSA_E_NOT_INITIALISED    |      Global initialization not successful
 *  VSA_E_NO_SPACE           |      Any resource allocation failed 
 *
 **********************************************************************/
VSA_RC
DLL_EXPORT VsaStartup( void )
{
    /* 
     * Comment: 
     * This dummy adapter does not need any further interfaces,
     * so return here VSA_OK.
     */
    /*--------------------------------------------------------------------*/
    /* The startup will be called process global                          */
    /* With this we say, we are ready to work                             */
    /*--------------------------------------------------------------------*/      
    if(bgInit == FALSE)
    {
        int rc = 0;
        /* CCQ_ON */
        rc = cl_init(CL_INIT_DEFAULT);
        if(rc) return VSA_E_LOAD_FAILED;
        /* CCQ_OFF */
        bgInit = TRUE;
    }
    return VSA_OK;
}


/**********************************************************************
 *  VsaGetConfig() 
 *
 *  Description:
 *     This call allows the SAP system to know which type of VSA was loaded and
 *     which parameters and other features it has. An AV vendor of a VSA
 *     can also define a default profile of initial and optional 
 *     parameters here. <nl>
 *     "REQ" (required) parameters are mandatory. <nl>
 *     "OPT" (optional) parameters are not mandatory.
 *
 *  Returncodes:
 *  VSA_OK                |         Success
 *  VSA_E_NOT_INITIALISED |         Global initialization not successful
 *  VSA_E_NO_SPACE        |         Any resource allocation failed 
 *  VSA_E_NULL_PARAMETER  |         NULL pointer provided
 *
 **********************************************************************/
VSA_RC 
DLL_EXPORT VsaGetConfig(PPVSA_CONFIG pp_config)
{
    size_t len = 0;

    /*
     * Comment:
     * This structure helps to define a list of parameters. This adapter
     * supports following of the VSA parameters.
     * The length is set to 0 because there are only BOOL parameters. In
     * case of Char parameters you have to set the length of the string.
     */
    static MY_INITPARAMS _initparams[] = {
        { VS_IP_INITDRIVERS            ,   VS_TYPE_CHAR   ,      0,     0}, 
        { VS_IP_INITDRIVERDIRECTORY    ,   VS_TYPE_CHAR   ,      0,     0},
        { VS_IP_INITTEMP_PATH          ,   VS_TYPE_CHAR   ,      0,     0}

    };

    static MY_OPTPARAMS _optparams[] = {
        { VS_OP_SCANBESTEFFORT         ,   VS_TYPE_BOOL   ,      0,     (void*)0},
        { VS_OP_SCANALLFILES           ,   VS_TYPE_BOOL   ,      0,     (void*)1},
        { VS_OP_SCANEXTRACT            ,   VS_TYPE_BOOL   ,      0,     (void*)1},
        { VS_OP_SCANEXTRACT_SIZE       ,   VS_TYPE_SIZE_T ,      0,     (void*)10485760},
        { VS_OP_SCANEXTRACT_DEPTH      ,   VS_TYPE_INT    ,      0,     (void*)8}
    };

    VSA_RC      rc      =   VSA_OK;
    int         x       =   0,
                ipar    =   sizeof(_initparams)/sizeof(_initparams[0]),
                opar    =   sizeof(_optparams)/sizeof(_optparams[0]);
 
    if(bgInit == FALSE)
        return VSA_E_NOT_INITIALISED; /* no successful VsaStartup */

    /*--------------------------------------------------------------------*/
    /* VSA_CONFIG allocation. Return a filled structure in the function   */
    /*--------------------------------------------------------------------*/
    if(pp_config == NULL)
        return VSA_E_NULL_PARAM; /* no handle */
    (*pp_config)                           = (PVSA_CONFIG)calloc(1, sizeof(VSA_CONFIG));
    if ((*pp_config) == NULL)
        CLEANUP(VSA_E_NO_SPACE);

    /* init params */
    (*pp_config)->pInitParams              = (PVSA_INITPARAMS)calloc(1, sizeof(VSA_INITPARAMS));
    if ((*pp_config)->pInitParams == NULL)
        CLEANUP(VSA_E_NO_SPACE);
    (*pp_config)->pInitParams->pInitParam  = (PVSA_INITPARAM)calloc(ipar, sizeof(VSA_INITPARAM));
    if ((*pp_config)->pInitParams->pInitParam == NULL)
        CLEANUP(VSA_E_NO_SPACE);
    
    /* option params */
    (*pp_config)->pOptParams               = (PVSA_OPTPARAMS)calloc(1, sizeof(VSA_OPTPARAMS));
    if ((*pp_config)->pOptParams == NULL)
        CLEANUP(VSA_E_NO_SPACE);
    (*pp_config)->pOptParams->pOptParam    = (PVSA_OPTPARAM)calloc(opar, sizeof(VSA_OPTPARAM));
    if ((*pp_config)->pOptParams->pOptParam == NULL)
        CLEANUP(VSA_E_NO_SPACE);


    (*pp_config)->pAdapterInfo             = (PVSA_ADAPTERINFO)calloc(1, sizeof(VSA_ADAPTERINFO));
    if ((*pp_config)->pAdapterInfo == NULL)
        CLEANUP(VSA_E_NO_SPACE);

    (*pp_config)->struct_size              =    sizeof(VSA_CONFIG);

    /*--------------------------------------------------------------------*/
    /* loop to fill the supported parameter structure                     */
    /*--------------------------------------------------------------------*/
    for(x=0; x<ipar;x++)
    {
       /*  ipar : number of entries in initparams
           x    : actual index
           we add parameters from the default structure and set this as default config structure 
            */
        switch(_initparams[x].tCode)
        {
        case VS_IP_INITDRIVERS:
           {
              PChar  _drivers = (PChar)malloc(CLAMAV_DRIVERS_LN +1);
               if(_drivers == NULL)
                  CLEANUP(VSA_E_NO_SPACE);
               strncpy((char*)_drivers,CLAMAV_DRIVERS,CLAMAV_DRIVERS_LN);
               _drivers[CLAMAV_DRIVERS_LN]=0;

               VSAddINITParameter ( (*pp_config)->pInitParams->pInitParam,
                                    (*pp_config)->pInitParams->usInitParams,
                                    _initparams[x].tCode,
                                    _initparams[x].tType,
                                    CLAMAV_DRIVERS_LN,
                                    (char*)_drivers
                                  );
           }
        break;
        case VS_IP_INITDRIVERDIRECTORY:
           {
               PChar  _dir = (PChar)malloc(DRIVER_DIRECTORY_LN+1);
               if(_dir == NULL)
                  CLEANUP(VSA_E_NO_SPACE);
               strncpy((char*)_dir,DRIVER_DIRECTORY,DRIVER_DIRECTORY_LN);
               _dir[DRIVER_DIRECTORY_LN]=0;

               VSAddINITParameter ( (*pp_config)->pInitParams->pInitParam,
                                    (*pp_config)->pInitParams->usInitParams,
                                    _initparams[x].tCode,
                                    _initparams[x].tType,
                                    DRIVER_DIRECTORY_LN,
                                    (char*)_dir
                                  );
           }
        break;
        case VS_IP_INITTEMP_PATH:
           {
              PChar  _tmpPath = NULL; /* CCQ_OFF */
               if( getenv("TMPDIR") == NULL)
#ifdef _WIN32
                  _tmpPath = (PChar)strdup(".");
#else
                  _tmpPath = (PChar)strdup("/tmp");
#endif
               else
                  _tmpPath = (PChar)strdup( getenv("TMPDIR") );

               VSAddINITParameter ( (*pp_config)->pInitParams->pInitParam,
                                    (*pp_config)->pInitParams->usInitParams,
                                    _initparams[x].tCode,
                                    _initparams[x].tType,
                                    strlen((const char*)_tmpPath),
                                    (PChar)_tmpPath
                                  );
           }/* CCQ_ON */
        break;
        default:
            VSAddINITParameter ( (*pp_config)->pInitParams->pInitParam,
                                 (*pp_config)->pInitParams->usInitParams,
                                 _initparams[x].tCode,
                                 _initparams[x].tType,
                                 _initparams[x].lLength,
                                 _initparams[x].pvValue
                               );
        break;
        }
    }

    for(x=0; x<opar;x++)
    {
       /*  opar : number of entries in optparams
           x    : actual index
           we add parameters from the default structure and set this as default config structure 
        */
            VSAddOPTParameter (  (*pp_config)->pOptParams->pOptParam,
                                 (*pp_config)->pOptParams->usOptParams,
                                 _optparams[x].tCode,
                                 _optparams[x].tType,
                                 _optparams[x].lLength,
                                 _optparams[x].pvValue
                              );        
     
    }

    /*--------------------------------------------------------------------*/
    /* set adapter features   */
    /*--------------------------------------------------------------------*/
    (*pp_config)->uiVsaActionFlags =     VSA_AP_SCAN;

    (*pp_config)->uiVsaScanFlags   =     VSA_SP_FILE;

    (*pp_config)->uiVsaEvtMsgFlags =     uigVS_SAP_ALL;
    /* No client I/O callback supported for this VSA version
     * Otherwise you should set here the callback flags for CIO support.
     */
    (*pp_config)->uiVsaCIOMsgFlags =     0;

    /*--------------------------------------------------------------------*/
    /* set adapter info constants                                         */
    /*--------------------------------------------------------------------*/
    (*pp_config)->pAdapterInfo->struct_size          = sizeof(VSA_ADAPTERINFO);
    (*pp_config)->pAdapterInfo->usVsiVersion         = VSI_VERSION;
    (*pp_config)->pAdapterInfo->usVsaMajVersion      = VSA_ADAPTER_MAJVER;
    (*pp_config)->pAdapterInfo->usVsaMinVersion      = VSA_ADAPTER_MINVER;
    (*pp_config)->pAdapterInfo->tAdapterID           = VS_AD_CLAM;
    (*pp_config)->pAdapterInfo->tThreadingModel      = VS_THREAD_FREE; 
    SETERRORTEXT((*pp_config)->pAdapterInfo->pszVendorInfo , VSA_VENDORINFO );
    SETERRORTEXT((*pp_config)->pAdapterInfo->pszAdapterName ,VSA_ADAPTERNAME );

cleanup:
    if (rc == VSA_OK)
    {
        return (VSA_OK);
    }
    else
    {
        freeVSA_CONFIG(pp_config);
        return (rc);
    }

} /* VsaGetConfig */


/**********************************************************************
 *  VsaInit()
 *
 *  Description:
 *     Initializes or creates a new scan engine instance.
 *     Then assigns all the vendor specific data structures from
 *     actual AV product to VSA data structure.<nl>
 *     "REQ" (required) parameters are mandatory. <nl>
 *     "OPT" (optional) parameters are not mandatory.
 *
 *  Returncodes:
 *  VSA_OK                   |      Success
 *  VSA_E_EXPIRED            |      Engine or driver expired
 *  VSA_E_NO_SPACE           |      Any resource allocation failed 
 *  VSA_E_LOAD_FAILED        |      Load failed (such as external process)
 *  VSA_E_BAD_EXPRESSION     |      Regular expression syntax is invalid
 *  VSA_E_NULL_PARAM         |      NULL pointer provided
 *  VSA_E_INVALID_PARAM      |      At least one parameter is invalid
 *  VSA_E_DRIVER_FAILED      |      At least one driver failed
 *  VSA_E_NOT_SUPPORTED      |      At least one param or object is not supported
 *  VSA_E_CIO_FAILED         |      Client I/O request action failed.
 *  VSA_E_NOT_INITIALISED    |      Global initialization not successful
 *  VSA_E_CBC_TERMINATED     |      Action was terminated during callback
 *
 **********************************************************************/
VSA_RC 
DLL_EXPORT VsaInit( const PVSA_CALLBACK p_callback, 
                    const PVSA_INITPARAMS p_initparams, 
                          PPVSA_INIT pp_init )
{
    VSA_RC                      rc      =   VSA_OK;
    Int                         i       =   0;
    unsigned int                iDrivers=   0;
    size_t                      len     =   0;
    PChar                      pDriverName;
    PChar                      szinitDrivers = NULL;
    INITDATA                   initConfig = {NULL,NULL,NULL};
    struct tm                 _utc_date = { 0,      /* hours                    */
                                            0,      /* daylight                 */
                                            0,      /* day of month (1-31)      */
                                            0,      /* minutes after hour (0-59)*/
                                            0,      /* month (0-11, january=0)  */
                                            0,      /* seconds after min (0-59) */
                                            0,      /* day of week, 0-6,Sun=0   */
                                            0,      /* day of year, 0-365,Jan1=0*/
                                            0};     /* year (current minus 1900 */
    /*   ----- clam param ---- */
    unsigned int dboptions = 0, sigs = 0;
    int ret = 0;
    struct cl_engine *engine = NULL;

    if(pp_init == NULL)
        return VSA_E_NULL_PARAM; /* no handle */

    if(bgInit == FALSE) 
    {
        SETERRORTEXT((*pp_init)->pszErrorText, "VsaStartup not yet called or not successful");
        (*pp_init)->iErrorRC = 5;
        return VSA_E_NOT_INITIALISED; /* no successful VsaStartup */
    }
    /* set date, e.g. 15.08.2009 */
    ENGINE_DATA(25,3,2011)

    /* Comment:
     * In the VsaInit function you should either connect/contact your
     * internal engine or allocate memory for the scan instance.
     * The structure VSA_INIT contains several flags which should help you
     * to pass your scan handle, internal license information, flags and so on to
     * the VsaScan function.
     */
    /* Initialize VSA_INIT structure */
    (*pp_init)          =   (PVSA_INIT)calloc(1,sizeof(VSA_INIT));
    if ((*pp_init) == NULL)
        CLEANUP(VSA_E_NO_SPACE);
        /* sizeof */
    (*pp_init)->struct_size = sizeof(VSA_INIT);

    /* Important hint:
     * Set hEngine with a value != NULL,
     * here you should set the HANDLE of your internal
     * engine instance.
     */ 
    (*pp_init)->uiViruses   = 1;
    (*pp_init)->uiExtensions= 0;
    (*pp_init)->uiIntRevNum = 0;
    (*pp_init)->uiSignature = 0xbbbbbbbb; /* add here an own magic */
    (*pp_init)->usDrivers   = 0;
    (*pp_init)->usEngineMajVersion = cl_retflevel();
    (*pp_init)->usEngineMinVersion = 0;
    SETSTRING( (*pp_init)->pszEngineVersionText, VSA_VERSION_STRING );
    /* set VSA_DRIVERINFO structure */
        (*pp_init)->iErrorRC   = 0;
    (*pp_init)->pDriver = (PVSA_DRIVERINFO)calloc(MAX_DRIVERS,sizeof(VSA_DRIVERINFO));
    if ((*pp_init)->pDriver == NULL)
        CLEANUP(VSA_E_NO_SPACE);

    rc = vsaSetInitConfig(p_initparams,&initConfig);
    if(rc) CLEANUP(rc);
    /* CCQ_OFF */
    engine = cl_engine_new( );
    if (engine == NULL)
    {
        SETERRORTEXT((*pp_init)->pszErrorText, "ClamAV engine initalization failed");
        CLEANUP(VSA_E_LOAD_FAILED);
    }
    ret = cl_load((const char*)initConfig.initdirectory->pvValue,engine,&sigs,dboptions);
    if(ret)
    {
       SETERRORTEXT((*pp_init)->pszErrorText, "ClamAV engine could not load db files");
       (*pp_init)->iErrorRC   = ret;
       cl_engine_free(engine);
       CLEANUP(VSA_E_DRIVER_FAILED);
    }
    /* CCQ_ON */
    if((ret = cl_engine_compile(engine)))
    {  /* CCQ_OFF */
       cl_engine_free(engine);
       (*pp_init)->iErrorRC   = ret;
       SETERRORTEXT((*pp_init)->pszErrorText, "ClamAV engine could not compile db files");
       CLEANUP(VSA_E_LOAD_FAILED);
    }
     /* convert date to calender date *//*CCQ_CLIB_LOCTIME_OK*/
    (*pp_init)->utcDate     = (time_t)cl_engine_get_num(engine,CL_ENGINE_DB_TIME,&ret);
    if(ret)
    {
        mktime(&_utc_date);
    }
    ret = cl_engine_set_str(engine, CL_ENGINE_TMPDIR, (const char*)initConfig.tmpdir->pvValue);
    (*pp_init)->hEngine     = (PVoid)engine;
    (*pp_init)->uiViruses   = sigs;
    /* CCQ_ON */
    /*
     * Comment: 
     * We allocate one driver info to demonstrate the use case.
     * The VSA specification does not dictate the usage of VSA_DRIVERINFO.
     * The information here will only be displayed at SAP internal applications,
     * but this information should help you later to anaylse with which pattern
     * files your engine runs.
     * This information also helps any customer to see which version of pattern
     * files are loaded.
     */    
    do {
       struct cl_cvd *p_driver;/* CCQ_OFF */
       szinitDrivers = (PChar)strdup((const char*)initConfig.drivers->pvValue);
       pDriverName = (PChar)strtok((char*)szinitDrivers,",");/*CCQ_FUNCTION_WITH_MEMORY_OK*/
       while(pDriverName!=NULL)
       {
          (*pp_init)->pDriver[i].pszName = (PChar)malloc(initConfig.initdirectory->lLength + strlen((const char*)pDriverName) + 3);
          if((*pp_init)->pDriver[i].pszName == NULL)
             CLEANUP(VSA_E_NO_SPACE);

          sprintf((char*)(*pp_init)->pDriver[i].pszName,"%s%s%s", (char*)initConfig.initdirectory->pvValue, DIR_SEP, pDriverName);
          if(getFileSize((*pp_init)->pDriver[i].pszName,NULL)) /* search for *.cvd */
          {   /* if *.cvd not found, then try it with *.cld */
              (*pp_init)->pDriver[i].pszName[(int)strlen((const char*)(*pp_init)->pDriver[i].pszName)-2] = 'l';
              if(getFileSize((*pp_init)->pDriver[i].pszName,NULL))
              { /* if not found, then not available */
                  (*pp_init)->pDriver[i].pszName = NULL;
              }
          }
          p_driver = NULL; /* initialise the structure */
          if( (*pp_init)->pDriver[i].pszName )
          {
              p_driver = cl_cvdhead((const char*)(*pp_init)->pDriver[i].pszName );
          }
          if(p_driver)
          {
              (*pp_init)->usDrivers++;
              (*pp_init)->pDriver[i].uiViruses = p_driver->sigs;
              (*pp_init)->pDriver[i].iDriverRC = 0;
              (*pp_init)->pDriver[i].utcDate   = p_driver->stime;
              (*pp_init)->pDriver[i].usDrvMajVersion  = (*pp_init)->usEngineMajVersion;
              (*pp_init)->pDriver[i].usDrvMinVersion  = p_driver->version;
              cl_cvdfree(p_driver);
          }
          else
          {
              /* convert date to calender date *//*CCQ_CLIB_LOCTIME_OK*/
              (*pp_init)->pDriver[i].utcDate = mktime(&_utc_date);
              (*pp_init)->pDriver[i].uiViruses        = 0;
              (*pp_init)->pDriver[i].uiVariants       = 0;  
              (*pp_init)->pDriver[i].iDriverRC        = -1;
              (*pp_init)->pDriver[i].usDrvMajVersion  = 1;
              (*pp_init)->pDriver[i].usDrvMinVersion  = 0;
          }
          (*pp_init)->pDriver[i].struct_size      = sizeof(VSA_DRIVERINFO);
          i++;
          pDriverName = (PChar)strtok(NULL,","); /*CCQ_FUNCTION_WITH_MEMORY_OK*/
       }       
    } while( pDriverName );/* CCQ_ON */

    /* check drivers */
    if( (*pp_init)->usDrivers < MIN_DRIVERS )
    {
       (*pp_init)->iErrorRC   = 9;
       SETERRORTEXT((*pp_init)->pszErrorText, "ClamAV engine drivers not fully loaded");
       CLEANUP(VSA_E_DRIVER_FAILED);
    }
cleanup:
    if(szinitDrivers) free(szinitDrivers);
    if (rc != VSA_OK)
    {
        if((*pp_init)->pszErrorText==NULL) SETSTRING( (*pp_init)->pszErrorText, cl_strerror(ret) );
        if (pp_init && (*pp_init) && (*pp_init)->iErrorRC == 0)
            freeVSA_INIT(pp_init);
    }
    else
    {   /* increase the ref. counter */
        lgRefCounter++;
        SETSTRING( (*pp_init)->pszErrorText, "No error" );
    }
    return (rc);
} /* VsaInit */



/**********************************************************************
 *  VsaScan()
 *
 *  Description:
 *     Performs the scan. Requires a valid instance handle VSA_INIT from
 *     <f VsaInit>. If the VSA should not perform any callback during the
 *     scan action, then the parameter pVsaScanparam can be set to NULL.
 *     The address of a handle to a <t VSA_SCANINFO> structure can optionally
 *     be provided. This means if "ppVsaScaninfo" is set to NULL, the VSA
 *     will not return any information about the scan action.
 *     This allows the caller to decide, if [only] callbacks [or \| and]
 *     [only] the VSA_SCANINFO structure should be returned. <nl>
 *     "REQ" (required) parameters are mandatory. <nl>
 *     "OPT" (optional) parameters are not mandatory.
 *
 *  Returncodes:
 *  Virus error codes (negative values):
 *  VSA_E_CLEAN_FAILED         |    Removing/replacing infection failed
 *  VSA_E_PATTERN_FOUND        |    Pattern was found
 *  VSA_E_MACRO_FOUND          |    Macro was found
 *  VSA_E_VIRUS_FOUND          |    Virus was found
 *  VSA_E_CLEAN_OK             |    The clean action was successful
 *  No error, no virus:
 *  VSA_OK                     |    Success
 *  Program error codes (positive values):
 *  VSA_E_NO_SPACE             |    Any resource allocation failed 
 *  VSA_E_NULL_PARAM           |    NULL pointer provided
 *  VSA_E_INVALID_PARAM        |    At least one parameter is invalid
 *  VSA_E_INVALID_HANDLE       |    The provided handle is invalid
 *  VSA_E_NOT_INITIALISED      |    The adapter was not successfully initialized
 *  VSA_E_NOT_SUPPORTED        |    At least one parameter or object is not supported
 *  VSA_E_INVALID_SCANOBJECT   |    See VSA_SCANPARAM, object is invalid
 *  VSA_E_CIO_FAILED           |    Client I/O failed. Scan could not be performed.
 *  VSA_E_SCAN_FAILED          |    The scan action failed
 *  VSA_E_NOT_SCANNED          |    At least one object was not scanned
 *  VSA_E_CBC_TERMINATED       |    Action was terminated during callback
 *
 **********************************************************************/
VSA_RC 
DLL_EXPORT VsaScan(
                 const PVSA_INIT      p_init, 
                 const PVSA_CALLBACK  p_callback,
                 const PVSA_SCANPARAM p_scanparam,
                 const PVSA_OPTPARAMS p_optparams, 
                 PPVSA_SCANINFO pp_scinfo
                )
{
    VSA_RC              rc              = VSA_OK;
    int                 clam_rc         = 0;
    VS_CALLRC           _vsa_rc         = VS_CB_OK;
    size_t              len             = 0;    
    VSA_EVENTCBFP       cfunc           = NULL;
    UInt                uiMsgFlag       = 0;
    VSA_USRDATA         pvUsrdata       = NULL;
    VSA_ENGINE          heng            = NULL;
    PChar               pszReason       = NULL;    
    PVSA_SCANERROR      p_scanerror     = NULL;
    PVSA_VIRUSINFO      p_virusinfo     = NULL;
    FILE                *_fp            = NULL;    
    const char          *virname = NULL;
    unsigned int        options = CL_SCAN_STDOPT;
    unsigned long int   scanned = 0;

    USRDATA             usrdata;
    memset(&usrdata,0,sizeof(USRDATA)); 
    /* --- removed some struct members with clamav 0.93 ---
    limits.maxfiles = 500;
    limits.maxreclevel = 8;
    limits.maxmailrec = 64;
    limits.maxratio = 250;
    limits.maxfilesize = 10485760;
    */
    
    if(bgInit == FALSE)
        CLEANUP(VSA_E_NOT_INITIALISED); /* no successful VsaStartup */   

    if (p_init == NULL || p_scanparam ==  NULL)
        CLEANUP(VSA_E_NULL_PARAM);

    /*--------------------------------------------------------------------*/
    /* check structure size to protect yourself against different versions*/
    /*--------------------------------------------------------------------*/
    if (p_scanparam->struct_size != sizeof(VSA_SCANPARAM))
        CLEANUP(VSA_E_INVALID_PARAM);

    /* check callback structure passed to me */
    rc = registerCallback(p_callback,&usrdata);
    if (rc)
        CLEANUP(rc);

    uiMsgFlag = usrdata.uiMsgFlags;
    pvUsrdata = usrdata.pvUsrdata;
    cfunc     = usrdata.pvFncptr;

    /*
     * allocate VSA_SCANINFO 
     */
    if (pp_scinfo != NULL) {
        (*pp_scinfo)         = (PVSA_SCANINFO)calloc(1,sizeof(VSA_SCANINFO));
        if ((*pp_scinfo) == NULL)
            CLEANUP(VSA_E_NO_SPACE);

        (*pp_scinfo)->struct_size= sizeof(VSA_SCANINFO);
        (*pp_scinfo)->uiJobID    = p_scanparam->uiJobID;
    }
    /*--------------------------------------------------------------------*/
    /* Check, if the action is VSA_AP_SCAN                                */
    /*--------------------------------------------------------------------*/
    if (p_scanparam->tActionCode != VSA_AP_SCAN)
    {
        pszReason = (PChar)"The VSA does not know this action!";
        CLEANUP(VSA_E_NOT_SUPPORTED);
    }
    /*--------------------------------------------------------------------*/
    /* Comment:                                                           */
    /* Set scan parameter configuration                                   */ 
    /*--------------------------------------------------------------------*/
    rc = vsaSetScanConfig(p_scanparam, p_optparams,&usrdata,(struct cl_engine*)p_init->hEngine);
    if (rc)
    {
        pszReason = (PChar)"At least one parameter is invalid!";
        CLEANUP(rc);
    }
    /*--------------------------------------------------------------------*/
    /* example callbacks to query whether we should start                 */
    /*--------------------------------------------------------------------*/
    _vsa_rc = CB_FUNC( VS_M_ABORTSCAN, (size_t)p_scanparam->uiJobID );
    if (_vsa_rc == VS_CB_NEXT || _vsa_rc == VS_CB_TERMINATE)
        CLEANUP(VSA_E_CBC_TERMINATED);

    _vsa_rc = CB_FUNC( VS_M_OBJECTFOUND, p_scanparam->pszObjectName );
    if (_vsa_rc == VS_CB_NEXT || _vsa_rc == VS_CB_TERMINATE)
        CLEANUP(VSA_E_CBC_TERMINATED);

    /*--------------------------------------------------------------------*/
    /* Comment:                                                           */
    /* Start here to process the different action types.                  */
    /* First we perform some plausi checks                                */
    /*--------------------------------------------------------------------*/
    /* 
    * Scan local file
    */
    switch(p_scanparam->tScanCode)    
    { 
    case VSA_SP_FILE:
        {
           rc = getFileSize(p_scanparam->pszObjectName,&usrdata.lObjectSize);
           if (rc) {
               pszReason = (PChar)"The file could not be opened!";
               CLEANUP( VSA_E_SCAN_FAILED );
           }           /* CCQ_OFF */
           clam_rc = cl_scanfile((const char*)p_scanparam->pszObjectName, 
                                 (const char**)&virname, 
                                 &scanned, 
                                 (const struct cl_engine *)p_init->hEngine, 
                                 options); 
                       /* CCQ_ON */
           if (clam_rc != CL_CLEAN)
           {
              pszReason = (PChar)cl_strerror(clam_rc);
              switch(clam_rc)
              {
              case CL_VIRUS: rc = VSA_E_VIRUS_FOUND; 
              break;
              default:       rc = VSA_E_SCAN_FAILED;
              pszReason = (PChar)"ClamAV engine with internal,unknown error.";
              break;
              }
           }
           else
           {
              rc = VSA_OK;
           }
        }          
    break;  
    default:
        CLEANUP(VSA_E_INVALID_SCANOBJECT);
    }

    /*
     * After the scan
     */
    if( clam_rc == CL_VIRUS )
    {
        /* found EICAR */
        p_virusinfo = (PVSA_VIRUSINFO)calloc(1,sizeof(VSA_VIRUSINFO));
        
        if(p_virusinfo== NULL)
            CLEANUP(VSA_E_NO_SPACE);
        
        p_virusinfo->struct_size = sizeof(VSA_VIRUSINFO);
        p_virusinfo->bRepairable = FALSE;
        p_virusinfo->lObjectSize = usrdata.lObjectSize;
        SETSTRING( p_virusinfo->pszFreeTextInfo , "No info available");
        SETSTRING( p_virusinfo->pszObjectName, p_scanparam->pszObjectName);
        SETSTRING( p_virusinfo->pszVirusName, virname);
        p_virusinfo->tActionType = VS_AT_NOACTION;
        p_virusinfo->tDetectType = VS_DT_KNOWNVIRUS;
        p_virusinfo->tVirusType  = VS_VT_VIRUS;
        p_virusinfo->uiVirusID   = 0;
        p_virusinfo->tObjectType = VS_OT_BINARY;

        SET_VSA_RC( VSA_E_VIRUS_FOUND );
        _vsa_rc = CB_FUNC( VS_M_VIRUS, p_virusinfo );
        if (_vsa_rc == VS_CB_NEXT || _vsa_rc == VS_CB_TERMINATE)
            CLEANUP(VSA_E_CBC_TERMINATED);

        if(pp_scinfo != NULL && (*pp_scinfo) != NULL)
        {
            (*pp_scinfo)->uiScanned++;
            (*pp_scinfo)->pVirusInfo = p_virusinfo;
            (*pp_scinfo)->uiInfections++;
        }
        else
        {   
            freevirusinfo(&p_virusinfo);
        }
    }
    else
    {  /*if( scanned > 0 ) *//* only if data was scanned */
       /*{*/
            if(pp_scinfo != NULL && (*pp_scinfo) != NULL)
            {
                (*pp_scinfo)->uiScanned++;
                (*pp_scinfo)->uiClean++;
            }
            SET_VSA_RC( VSA_OK );
            /* no virus found */
            _vsa_rc = CB_FUNC( VS_M_CLEAN, (size_t)p_scanparam->uiJobID );
            if (_vsa_rc == VS_CB_NEXT || _vsa_rc == VS_CB_TERMINATE)
                CLEANUP(VSA_E_CBC_TERMINATED);
        /*}*/
        /*else
        {
            rc = VSA_E_NOT_SCANNED;                         
        }*/
    }
    /* Exception handling */
cleanup:
    if (_fp != NULL)
    {
        fclose(_fp);
        _fp = NULL;
    }
    switch(rc)
    {
    case VSA_E_NOT_SUPPORTED:
        SET_VSA_RC( VSA_E_NOT_SUPPORTED );
        setScanError(p_scanparam->uiJobID,
                     p_scanparam->pszObjectName,
                     usrdata.lObjectSize,
                     clam_rc,
                     pszReason,
                     &p_scanerror
                     );
        _vsa_rc = CB_FUNC( VS_M_ERROR, pszReason );
        if (_vsa_rc == VS_CB_NEXT || _vsa_rc == VS_CB_TERMINATE)
            SET_VSA_RC(VSA_E_CBC_TERMINATED);
    break;
    case VSA_E_NOT_SCANNED:
        SET_VSA_RC( VSA_E_NOT_SCANNED );
        setScanError(p_scanparam->uiJobID,
                     p_scanparam->pszObjectName,
                     usrdata.lObjectSize,
                     clam_rc,
                     (PChar)"This object was not scanned",
                     &p_scanerror
                     );
        _vsa_rc = CB_FUNC( VS_M_NOTSCANNED, p_scanerror );
        if (_vsa_rc == VS_CB_NEXT || _vsa_rc == VS_CB_TERMINATE)
            SET_VSA_RC(VSA_E_CBC_TERMINATED);
    break;
    case VSA_E_SCAN_FAILED:
        SET_VSA_RC( VSA_E_SCAN_FAILED );
        setScanError(p_scanparam->uiJobID,
                     p_scanparam->pszObjectName,
                     usrdata.lObjectSize,
                     clam_rc,
                     pszReason,
                     &p_scanerror
                     );
        _vsa_rc = CB_FUNC( VS_M_ERROR, pszReason );
        if (_vsa_rc == VS_CB_NEXT || _vsa_rc == VS_CB_TERMINATE)
            SET_VSA_RC(VSA_E_CBC_TERMINATED);
    break;
    default:
    break;    
    }
    if( rc == VSA_E_NOT_SCANNED   ||
        rc == VSA_E_NOT_SUPPORTED ||
        rc == VSA_E_SCAN_FAILED)
    {     
        if(pp_scinfo != NULL && (*pp_scinfo) != NULL)
        {
            (*pp_scinfo)->pScanError    =   p_scanerror;
            (*pp_scinfo)->uiScanErrors++;
            (*pp_scinfo)->uiNotScanned++;
        }
        else
        {
            freescanerror(&p_scanerror);
        }
    }
    if (rc == 0)
        rc = usrdata.vsa_rc; /* set now the saved RC to return value */
    if (rc == 0)
        return VSA_OK;
    else if (usrdata.vsa_rc == 0) /* program error , otherwise the Virus error would be 0 */ 
    {
        VsaReleaseScan(pp_scinfo);
    }
    return (rc);
} /* VsaScan */


/**********************************************************************
 *  VsaReleaseScan()
 *
 *  Description:
 *     Release the dynamically allocated struture VSA_SCANINFO. The address of the
 *     of the handle is required but the handle can also point to NULL. <nl>
 *     "REQ" (required) parameters are mandatory. <nl>
 *     "OPT" (optional) parameters are not mandatory.
 *  
 *  Returncodes:
 *  VSA_OK                   |      Success
 *  VSA_E_NOT_INITIALISED    |      Global initialization not successful
 *  VSA_E_NULL_PARAM         |      NULL pointer provided
 *
 **********************************************************************/
VSA_RC 
DLL_EXPORT VsaReleaseScan(PPVSA_SCANINFO ppscinfo)
{
    UInt i = 0;

    if(bgInit == FALSE)
        return VSA_E_NOT_INITIALISED; /* no successful VsaStartup */
    if(ppscinfo == NULL)
        return VSA_E_NULL_PARAM;
    /*--------------------------------------------------------------------*/
    /* free the VSA_SCANINFO structure and sub structure VSA_SCANINFO     */
    /*--------------------------------------------------------------------*/
    if( ppscinfo != NULL && (*ppscinfo) != NULL && (*ppscinfo)->pVirusInfo != NULL )
    {
        for(i=0 ; i < (UInt)((*ppscinfo)->uiInfections); i++)
        {
            freevirusinfo2(&((*ppscinfo)->pVirusInfo[i]));
        }
        free((*ppscinfo)->pVirusInfo);
        (*ppscinfo)->pVirusInfo = NULL;
    }
    /*--------------------------------------------------------------------*/
    /* free VSA_SCANERROR                                                 */
    /*--------------------------------------------------------------------*/
    if( ppscinfo != NULL && (*ppscinfo) != NULL && (*ppscinfo)->pScanError != NULL)
    {
        for(i=0 ; i < (UInt)((*ppscinfo)->uiScanErrors); i++)
        {
            freescanerror2(&((*ppscinfo)->pScanError[i]));
        }
        free((*ppscinfo)->pScanError);
        (*ppscinfo)->pScanError = NULL;
    }
    /*--------------------------------------------------------------------*/
    /* free the pointer and reset it, so that the callee receives NULL    */
    /*--------------------------------------------------------------------*/
    if( (*ppscinfo) != NULL) {
        free((*ppscinfo));
        (*ppscinfo) = NULL;
    }        
    return (VSA_OK);

} /* VsaReleaseScan */


/**********************************************************************
 *  VsaEnd()
 *
 *  Description:
 *     Closes the engine instance. Releases also VSA_CONFIG allocated
 *     by VsaGetConfig - assumes that the actual job is done there. <nl>
 *     "REQ" (required) parameters are mandatory. <nl>
 *     "OPT" (optional) parameters are not mandatory.
 *
 *  Returncodes:
 *  VSA_OK                     |    Success
 *  VSA_E_NULL_PARAM           |    NULL pointer provided
 *  VSA_E_NOT_INITIALISED      |    Global initialization not successful
 *  VSA_E_END_FAILED           |    The AV engine could not be closed
 *  VSA_E_IN_PROGRESS          |    Any thread is still running
 *
 **********************************************************************/
VSA_RC 
DLL_EXPORT VsaEnd(PPVSA_INIT pp_init, PPVSA_CONFIG pp_config)
{ 
    VSA_RC        rc    = VSA_OK;

    if(bgInit == FALSE)
        CLEANUP( VSA_E_NOT_INITIALISED ); /* no successful VsaStartup */
    if(pp_init == NULL || pp_config == NULL)
        CLEANUP( VSA_E_NULL_PARAM );
    /*--------------------------------------------------------------------*/
    /* free VSA_INIT                                                      */
    /*--------------------------------------------------------------------*/
    if (pp_init != NULL && (*pp_init) != NULL)
    {
        if((*pp_init)->hEngine)  /* CCQ_OFF */
           cl_engine_free((struct cl_engine *)(*pp_init)->hEngine);
        freeVSA_INIT(pp_init);   /* CCQ_ON */
    }   
        
    freeVSA_CONFIG(pp_config);

cleanup:

    return (rc);
} /* VsaEnd */


/**********************************************************************
 *  VsaCleanup()
 *
 *  Description:
 *     Global cleanup for the adapter. This function will be called at last
 *     before unloading the VSA.
 *
 *  Returncodes:
 *  VSA_OK                     |    Success
 *  VSA_E_NOT_INITIALISED      |    Global initialization not successful
 *  VSA_E_IN_PROGRESS          |    Any thread is still running
 *
 **********************************************************************/
VSA_RC 
DLL_EXPORT VsaCleanup( void )
{
    /* 
     * cleanup
     */
    if(bgInit == FALSE)
        return VSA_E_NOT_INITIALISED; /* no successful VsaStartup */
    
    if(lgRefCounter != (size_t)0)
        return VSA_E_IN_PROGRESS;     /* any instance is still active */
    /*--------------------------------------------------------------------*/
    /* The cleanup will be called process global                          */
    /*--------------------------------------------------------------------*/    
    bgInit = FALSE;
    return VSA_OK;
} /* VsaCleanup */


/**********************************************************************
 **********************************************************************
 *
 *
 *  PRIVATE helper/auxiliary functions, here only for this prototype implementation
 *  If you would built an adapter, then you can decide by yourself, 
 *  which parts you want have here or not.
 *
 *
 **********************************************************************
 **********************************************************************/
static VSA_RC setScanError(UInt            uiJobID,
                           PChar           pszObjectName,
                           size_t          lObjectSize,
                           Int             iErrorRC,
                           PChar           pszErrorText,
                           PPVSA_SCANERROR pp_scanerror)
{

    size_t len = 0;
    VSA_RC rc  = VSA_OK;

    if (pp_scanerror == NULL)
        return VSA_E_NULL_PARAM;

   (*pp_scanerror) = (PVSA_SCANERROR)calloc(1,sizeof(VSA_SCANERROR));
        
    if( (*pp_scanerror) == NULL)
        return VSA_E_NO_SPACE;
  
    (*pp_scanerror)->struct_size = sizeof(VSA_SCANERROR);
    if(pszObjectName)
        SETSTRING( (*pp_scanerror)->pszObjectName, pszObjectName );
    (*pp_scanerror)->lObjectSize = lObjectSize;
    (*pp_scanerror)->iErrorRC    = iErrorRC;
    if(pszErrorText)
        SETSTRING( (*pp_scanerror)->pszErrorText, pszErrorText );
    (*pp_scanerror)->uiJobID     = uiJobID;
    
cleanup:
    return rc;
} /* setScanError */


static VSA_RC vsaSetScanConfig(VSA_SCANPARAM *p_scanparam,VSA_OPTPARAMS *p_optparams, USRDATA *usrdata, struct cl_engine *engine)
{
    VSA_RC     rc        = VSA_OK;
    Int        i         = 0,
               arraysize = (p_optparams?p_optparams->usOptParams:0);
    /*
     * set the optional setting
     */
    for (i=0; i<arraysize; i++)
    {
        switch (p_optparams->pOptParam[i].tCode)
        {/* SCANBESTEFFORT should provide a scan with "best effort" of the engine.
          * this interface of Sym. Scan Engine does not allow to set such parameters.
          * but at least the adapter must know the parameter type!
          */
        case VS_OP_SCANBESTEFFORT:
             if ((p_optparams->pOptParam[i].pvValue)!=NULL)
                 usrdata->bScanBestEffort = TRUE;
        break;
        case VS_OP_SCANALLFILES:
             if ((p_optparams->pOptParam[i].pvValue)!=NULL)
                 usrdata->bScanAllFiles = TRUE;
        break;
        case VS_OP_SCANEXTRACT_SIZE:
             if ((p_optparams->pOptParam[i].pvValue)!=NULL)
                 cl_engine_set_num(engine, CL_ENGINE_MAX_FILESIZE, (long long) ((size_t)p_optparams->pOptParam[i].pvValue));
        break;
        case VS_OP_SCANEXTRACT_DEPTH:
             if ((p_optparams->pOptParam[i].pvValue)!=NULL)
                cl_engine_set_num(engine, CL_ENGINE_MAX_RECURSION, (long long) ((size_t)p_optparams->pOptParam[i].pvValue));
        break;
        default:
        break;       
        }
    }
   
    return rc;
}

static VSA_RC vsaSetInitConfig(VSA_INITPARAMS *p_intparams,
                               INITDATA *usrdata
                               )
{
    VSA_RC     rc        = VSA_OK;
    Int        i         = 0,
               arraysize = (p_intparams?p_intparams->usInitParams:0);
    /*
     * set the initial setting
     */
    for (i=0; i<arraysize; i++)
    {
        switch (p_intparams->pInitParam[i].tCode)
        {
        case VS_IP_INITDRIVERDIRECTORY:
           usrdata->initdirectory = &p_intparams->pInitParam[i];
        break;
        case VS_IP_INITDRIVERS:
           usrdata->drivers = &p_intparams->pInitParam[i];
        break;
        case VS_IP_INITTEMP_PATH:
           usrdata->tmpdir = &p_intparams->pInitParam[i];
        break;
        default:
             return VSA_E_NOT_SUPPORTED;        
        }
    }
    return rc;
}

static VSA_RC getFileSize(Char *pszfilename, size_t *size)
{
      VSA_RC rc = VSA_OK;

      struct stat     buf;
        if (pszfilename == NULL)
            return VSA_E_NULL_PARAM;
        
        if (0!=(stat((const char*)pszfilename,&buf)))
            return VSA_E_SCAN_FAILED;
      
        if(size!=NULL)
        (*size)  = buf.st_size;

        return rc;
}

static VSA_RC registerCallback(VSA_CALLBACK *p_callback, USRDATA *usrdata)
{

    VSA_RC rc = VSA_OK;

    /* initialize the helper structure USRDATA */
    memset(usrdata,0,sizeof(USRDATA));

    if (p_callback != NULL)
    {
        if (p_callback->struct_size != sizeof(VSA_CALLBACK))
            CLEANUP(VSA_E_INVALID_PARAM);
        if (p_callback->pEventCBFP != NULL && p_callback->uiEventMsgFlags != 0)
        {
            if (uigVS_SAP_ALL > (p_callback->uiEventMsgFlags) &&  /* only known messages */
                uigVS_SAP_ALL & (p_callback->uiEventMsgFlags))
            {
                usrdata->uiMsgFlags = p_callback->uiEventMsgFlags;
            }
            else if ( (p_callback->uiEventMsgFlags) == VS_M_ALL )
            {
                usrdata->uiMsgFlags = uigVS_SAP_ALL;     /* constant for all msg */
                
            }
            else
            {
                CLEANUP(VSA_E_NOT_SUPPORTED);            /* wished msg not supported */
            }
            usrdata->pvFncptr   = p_callback->pEventCBFP;            
            usrdata->pvUsrdata  = p_callback->pvUsrData;
        }
        if ( p_callback->pEventCBFP       == NULL ||  
             p_callback->uiEventMsgFlags == 0
            )           
        {

            p_callback->pEventCBFP        = NULL;
            p_callback->uiEventMsgFlags   = 0;
                  /* may somebody forgot this feature to set MsgFlags?
                   * anyway you want handle this (trace or error), i deactivate
                   * here the CALLBACK mechanism
                   */
        }
    }       

cleanup:
    return (rc);
}

static void freeVSA_INIT(VSA_INIT **pp_init)
{
  if(pp_init != NULL && (*pp_init) != NULL)
  {
    /* cleanup the driver structures */
    if ((*pp_init) != NULL && (*pp_init)->pDriver != NULL){
        int i = 0;
        for (i=0; i<(*pp_init)->usDrivers; i++)
        {
            if ((*pp_init) != NULL && (*pp_init)->pDriver != NULL && (*pp_init)->pDriver[i].pszName != NULL){
                free((*pp_init)->pDriver[i].pszName);   
                (*pp_init)->pDriver[i].pszName = NULL;
            }
        }
        free((*pp_init)->pDriver);
        (*pp_init)->pDriver = NULL;
    }    
    if ((*pp_init) != NULL && (*pp_init)->pszEngineVersionText != NULL)
        free((*pp_init)->pszEngineVersionText);
    if ((*pp_init) != NULL && (*pp_init)->pszErrorText != NULL)
        free((*pp_init)->pszErrorText);

    /* free the rest of the p_init */
    if ((*pp_init) != NULL) {
        free((*pp_init));    
    }
    /* set now to NULL */
    (*pp_init) = NULL;

    /* dec. the ref counter */
    if(lgRefCounter > (size_t)0)
        lgRefCounter--;
  }
}

static void freeVSA_CONFIG(VSA_CONFIG **pp_config)
{
  if(pp_config != NULL && (*pp_config) != NULL)
  {
    if ((*pp_config)->pInitParams != NULL && (*pp_config)->pInitParams->pInitParam != NULL)
    {
        int i=0;
        for(i=0; i<(*pp_config)->pInitParams->usInitParams; i++)
        {
            if ((*pp_config)->pInitParams->pInitParam[i].tType == VS_TYPE_CHAR &&
                (*pp_config)->pInitParams->pInitParam[i].pvValue != NULL)
                    free(((*pp_config)->pInitParams->pInitParam[i].pvValue));
        }
        free((*pp_config)->pInitParams->pInitParam);
    }
    if ((*pp_config)->pInitParams != NULL)
        free((*pp_config)->pInitParams);
    if ((*pp_config)->pOptParams != NULL && (*pp_config)->pOptParams->pOptParam != NULL)
    {
        int i=0;
        for(i=0; i<(*pp_config)->pOptParams->usOptParams; i++)
        {
            if ((*pp_config)->pOptParams->pOptParam[i].tType == VS_TYPE_CHAR &&
                (*pp_config)->pOptParams->pOptParam[i].pvValue != NULL)
                    free((*pp_config)->pOptParams->pOptParam[i].pvValue);
        }
        free((*pp_config)->pOptParams->pOptParam);
    }
    if ((*pp_config)->pOptParams != NULL)
        free((*pp_config)->pOptParams);

    if ((*pp_config)->pAdapterInfo != NULL )
    {
        if((*pp_config)->pAdapterInfo->pszVendorInfo != NULL)
          free((*pp_config)->pAdapterInfo->pszVendorInfo);
        if((*pp_config)->pAdapterInfo->pszAdapterName != NULL)
          free((*pp_config)->pAdapterInfo->pszAdapterName);
        /* free adapterinfo structure */
        free((*pp_config)->pAdapterInfo);
        (*pp_config)->pAdapterInfo = NULL;
    }
    
    free(*pp_config);
    /* set now to NULL */
    (*pp_config) = NULL;
  }
}

static void freevirusinfo(VSA_VIRUSINFO **vsainf)
{
  if(vsainf != NULL && (*vsainf) != NULL) {
    if ((*vsainf)->pszVirusName != NULL) {
        free((*vsainf)->pszVirusName);
        (*vsainf)->pszVirusName = NULL;
    }
    if ((*vsainf)->pszObjectName != NULL){
        free((*vsainf)->pszObjectName);
        (*vsainf)->pszObjectName = NULL;
    }
    if ((*vsainf)->pszFreeTextInfo != NULL){
        free((*vsainf)->pszFreeTextInfo);
        (*vsainf)->pszFreeTextInfo = NULL;
    } 
    if ((*vsainf) != NULL) {
        free((*vsainf));
        (*vsainf) = NULL;
    }
  }
}

static void freevirusinfo2(VSA_VIRUSINFO *vsainf)
{
  if(vsainf != NULL ) {
    if ((*vsainf).pszVirusName != NULL) {
        free((*vsainf).pszVirusName);
        (*vsainf).pszVirusName = NULL;
    }
    if ((*vsainf).pszObjectName != NULL){
        free((*vsainf).pszObjectName);
        (*vsainf).pszObjectName = NULL;
    }
    if ((*vsainf).pszFreeTextInfo != NULL){
        free((*vsainf).pszFreeTextInfo);
        (*vsainf).pszFreeTextInfo = NULL;
    }    
  }
}

static void freescanerror(VSA_SCANERROR **vsierror)
{
  if (vsierror != NULL && (*vsierror) != NULL) {
    if ((*vsierror)->pszObjectName != NULL) {
        free((*vsierror)->pszObjectName);
        (*vsierror)->pszObjectName = NULL;
    }
    if ((*vsierror)->pszErrorText!= NULL) {
        free((*vsierror)->pszErrorText);
        (*vsierror)->pszErrorText= NULL;
    }
    if ((*vsierror) != NULL) {
        free((*vsierror));
        (*vsierror) = NULL;
    }
  }
}
static void freescanerror2(VSA_SCANERROR *vsierror)
{
  if (vsierror != NULL ) {
    if ((*vsierror).pszObjectName != NULL) {
        free((*vsierror).pszObjectName);
        (*vsierror).pszObjectName = NULL;
    }
    if ((*vsierror).pszErrorText!= NULL) {
        free((*vsierror).pszErrorText);
        (*vsierror).pszErrorText= NULL;
    }

  }
}
